# Princeton University licenses this file to You under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.


# *********************************************  ObjectiveMechanism ****************************************************

"""

Contents
--------

  * `ObjectiveMechanism_Overview`
  * `ObjectiveMechanism_Creation`
      - `ObjectiveMechanism_Monitor`
  * `ObjectiveMechanism_Structure`
        - `ObjectiveMechanism_Input`
        - `ObjectiveMechanism_Function`
        - `ObjectiveMechanism_Output`
  * `ObjectiveMechanism_Execution`
  * `ObjectiveMechanism_Examples`
  * `ObjectiveMechanism_Class_Reference`

.. _ObjectiveMechanism_Overview:

Overview
--------

An ObjectiveMechanism is a `ProcessingMechanism <ProcessingMechanism>` that monitors the `OutputPorts <OutputPort>`
of one or more other ProcessingMechanisms specified in its `monitor <ObjectiveMechanism.monitor>` attribute,
and evaluates them using its `function <ObjectiveMechanism.function>`. The result of the evaluation is placed in the
ObjectiveMechanism's *OUTCOME* (`primary <OutputPort_Primary>`) OutputPort.  ObjectiveMechanisms are typically used
closely with (and often created automatically by) `ModulatoryMechanisms <ModulatoryMechanism>`.

.. _ObjectiveMechanism_Creation:

Creating an ObjectiveMechanism
------------------------------

ObjectiveMechanisms are often created automatically when other PsyNeuLink components are created (in particular,
ModulatoryMechanisms, such as `LearningMechanisms <LearningMechanism_Creation>` and
`ControlMechanisms <ControlMechanism_Creation>`).  An ObjectiveMechanism can also be created directly by calling its
constructor.  The primary attribute used to define an ObjectiveMechanism is its `monitor <ObjectiveMechanism.monitor>`
attribute, that is specified using the corresponding argument of its constructor as described below.

.. _ObjectiveMechanism_Monitor:

*Monitor*
~~~~~~~~~

COMMENT:
FOR DEVELOPERS:
    The monitor argument is in effect an alias to the input_ports argument
    of the constructor for a Mechanism;  it is simply assigned to input_port in the __init__ method
    and the specifications are handled by an override of the Mechanism's _instantiate_input_ports method.
    The monitor property returns the OutputPorts that project to the Mechanism's InputPorts
COMMENT

The **monitor** argument of an ObjectiveMechanism's constructor specifies the `OutputPorts <OutputPort>` it monitors.
This takes the place of the **input_ports** argument used by most other forms of `Mechanism <Mechanism>`, and is used
by the ObjectiveMechanism to create an `InputPort` for each OutputPort it monitors, along with a `MappingProjection`
from the OutputPort to that InputPort.  The **monitor** argument takes a list of items that can
include any of the `forms of specification <InputPort_Specification>` used in a standard **input_ports** argument.
For the **monitor** argument, this is usually a list of OutputPorts to be monitored.  However, as with a standard
**input_ports** argument, the **monitor** argument can include Mechanisms (in which case their `primary OutputPort
<OutputPort_Primary>` is used) or the `InputPort(s) <InputPort>` of other Mechanisms (in which case the
ObjectiveMechanism will be assigned Projections from all of the OutputPorts that project to the specified InputPort
-- that is, it will `shadow their inputs <InputPort_Shadow_Inputs>`). Items in the *monitor* argument can also be
used to specify attributes of the InputPort and/or MappingProjection(s) to it, that the ObjectiveMechanism creates to
monitor the specified OutputPort.  In general, the `value <OutputPort.value>` of each specified OutputPort determines
the format of the `variable <InputPort.variable>` of the InputPort that is created for it by the ObjectiveMechanism.
However, this can be overridden using the ObjectiveMechanism's `default_variable <Component_Variable>`
or `input_shapes <Mechanism_Base.input_shapes>` attributes (see `Mechanism InputPort specification
<Mechanism_InputPort_Specification>`), or by specifying a Projection from the OutputPort to the InputPort (see
`Input Source Specification <InputPort_Projection_Source_Specification>`). If an item in the
**monitor** argument specifies an InputPort for the ObjectiveMechanism, but not the OutputPort to
be monitored, the InputPort is created but will be ignored until an OutputPort (and MappingProjection from it) are
specified for that InputPort.

COMMENT:
FIX 8/27/19 [JDC]:
ADD DISCUSSION OF monitor_weights_and_exponents ARGUMENT HERE
.. _ObjectiveMechanism_Monitor_Weights_and_Exponents:
COMMENT

COMMENT:
Note that some forms of specification may depend on specifications made for the OutputPort referenced, the Mechanism
to which it belongs, and/or the Process or System to which that Mechanism belongs. These interactions (and the
precedence afforded to each) are described below.

  * **OutputPort** -- a reference to the `OutputPort <OutputPort>` of a Mechanism;  this creates an InputPort
    with a `variable <InputPort.variable>` that matches the format of the `value <OutputPort.value>` of the
    specified OutputPort, and a `MappingProjection` between them using an *IDENTITY_MATRIX*.

      TBI
      Note that an outputPort can be *excluded* from being monitored by assigning `None` as the value of its
      `monitoring_status` attribute.  This specification takes precedence over any others;  that is, it suppresses
      monitoring of that OutputPort, irrespective of any other specifications that might otherwise apply to that
      OutputPort, including those described below.
  ..
  TBI
  * **Mechanism** -- by default, the Mechanism's `primary OutputPort <OutputPort_Primary>` is used.  However,
    if the Mechanism has any OutputPorts specified in its `monitored_ports` attribute, those are used (except for
    any that specify `None` as their `monitoring_status`). This specification takes precedence over any of the other
    types listed below:  if it is `None`, then none of that Mechanism's OutputPorts are monitored; if it
    specifies OutputPorts to be monitored, those are monitored even if they do not satisfy any of the conditions
    described in the specifications below.

TBI FOR COMPOSITION:
The OutputPorts monitored by the ObjectiveMechanism are listed in its `monitor <ObjectiveMechanism.monitor>`
attribute.  When an ObjectiveMechanism is created by a `ControlMechanism`, or a `System` for its `controller
<System.controller>`, these may pass a set of OutputPorts to be monitored to the ObjectiveMechanism.  A
ControlMechanism passes OutputPort specifications listed in its **objective_mechanism** argument (see
`ControlMechanism_ObjectiveMechanism`), and a System passes any listed in its **monitor_for_control** argument (see
`System_Control_Specification`).
COMMENT

COMMENT:
FIX 8/27/19 [JDC]:
GENERALIZE TO ControlMechanism
COMMENT

The OutputPorts monitored by the ObjectiveMechanism are listed in its `monitor <ObjectiveMechanism.monitor>` attribute.
When an ObjectiveMechanism is created by a `ControlMechanism`, these may pass a set of OutputPorts to be monitored to
the ObjectiveMechanism.  A ControlMechanism passes OutputPort specifications listed in its **objective_mechanism**
argument (see `ControlMechanism_ObjectiveMechanism`).

.. _ObjectiveMechanism_Structure:


Structure
---------

.. _ObjectiveMechanism_Input:

*Input*
~~~~~~~

An ObjectiveMechanism has one `InputPort <InputPort>` for each of the OutputPorts specified in its
**monitor** argument (see `ObjectiveMechanism_Monitor`). Each InputPort receives a `MappingProjection` from the
corresponding OutputPort, the values of which are used by the ObjectiveMechanism's
`function <ObjectiveMechanism.function>` to generate the value of its *OUTCOME* `OutputPort
<ObjectiveMechanism_Output>`.  The InputPorts are listed in the ObjectiveMechanism's `input_ports
<ObjectiveMechanism.input_ports>` attribute, and the monitored OutputPorts from which they receive projections are
listed in the same order its `monitor <ObjectiveMechanism.monitor>` attribute.

COMMENT:
  FIX: Shadowing inputs may generate an exception to this, if the shadowed InputPort receives more than one projection.
       In that case, there will be more than on OutputPort listed in monitor for the corresponding input_port
COMMENT

By default, the format of the `variable <InputPort.variable>` for each InputPort is determined by the `value
<OutputPort.value>` of the monitored OutputPort(s) to which it corresponds.  However, if either the
**default_variable** or **input_shapes** argument is specified in an Objective Mechanism's constructor, or a `variable
<InputPort.variable>` is `specified for an InputPort <InputPort_Specification>` for one or more of the items in
its **monitor** argument, then that is used as the format for the corresponding InputPort(s).  This can be used to
transform the `value <OutputPort.value>` of a monitored OutputPort into different form for the `variable
<InputPort.variable>` of the InputPort (see the `first example <ObjectiveMechanism_Monitor_Examples>` below).

If the weight and/or exponent is specified for ay item in the **monitor** argument of the ObjectiveMechanism's
constructor, it is assigned to the corresponding InputPort.  If the ObjectiveMechanism's `function
<ObjectiveMechanism.function>` implements a weights and/or exponents attribute, the values specified is assigned to
the corresponding attribute, and applied to the `value <InputPort.value>` of the InputPort before it is combined
with that of the other InputPorts to generate the ObjectiveMechanism's `output <ObjectiveMechanism_Output>`.


.. _ObjectiveMechanism_Function:

*Function*
~~~~~~~~~~

The ObjectiveMechanism's `function <ObjectiveMechanism.function>` uses the values of its `input_ports
<ObjectiveMechanism.input_ports>` to compute an `objective (or "loss") function
<https://en.wikipedia.org/wiki/Loss_function>`_, that is assigned as the value of its *OUTCOME* `OutputPort
<ObjectiveMechanism_Output>`.  By default, it uses a `LinearCombination` function to sum the values of the values of
the items in its `variable <Mechanism_Base.variable>`. However, by assigning values to the 'weight
<InputPort.weight>` and/or 'exponent <InputPort.exponent>` attributes of the corresponding InputPorts,
it can be configured to calculate differences, ratios,  etc. (see `example
<ObjectiveMechanism_Weights_and_Exponents_Example>` below).  The `function <ObjectiveMechanism.function>`  can also
be replaced with any `TransformFunction <Transformfunctions>`, or any python function that takes an 2d array as
its input (with a number of items in axis 0 equal to the number of the ObjectiveMechanism's InputPorts), and generates
a 1d array as its result. If it implements :keyword:`weight` and/or :keyword:`exponent` attributes, those are assigned
from `weight <InputPort.weight>` and `exponent <InputPort.exponent>` attributes of its `input_ports
<ObjectiveMechanism.input_ports>` (also listed in the `monitor_weights_and_exponents
<ObjectiveMechanism.monitor_weights_and_exponents>` attribute);  otherwise, they are ignored.

.. _ObjectiveMechanism_Output:

*Output*
~~~~~~~~

The `primary OutputPort <OutputPort_Primary>` of an Objective mechanism is a 1d array, named *OUTCOME*, that is the
result of its `function <ObjectiveMechanism.function>` (as described above).


.. _ObjectiveMechanism_Execution:

Execution
---------

When an ObjectiveMechanism is executed, it updates its input_ports with the values of the OutputPorts listed in
its `monitor <ObjectiveMechanism.monitor>` attribute, and then uses its `function <ObjectiveMechanism.function>`
to evaluate these.  The result is assigned as to its `value <Mechanism_Base.value>` attribute as the value of its
*OUTCOME* (`primary <OutputPort_Primary>`) OutputPort.

.. _ObjectiveMechanism_Examples:

Examples
--------

.. _ObjectiveMechanism_Monitor_Examples:

*Specifying* the **variable** for the InputPorts of an ObjectiveMechanism

This can be useful in some situations, and there are several ways it can be done. For example, for
`Reinforcement Learning <Reinforcement>`, an ObjectiveMechanism is used to monitor the rewards by an action selection
Mechanism (and used by a LearningMechanism to improve those predictions).  However, whereas the action selection
Mechanism generates a vector indicating the reward predicted by the selected action, the ObjectiveMechanism for RL
simply needs to know the magnitude of the reward predicted, irrespective of the action taken;  that is, it
simply requires a single scalar value indicating the magnitude of the predicted reward.  Thus, the vector of
action-related reward predictions needs to be condensed to a single predicted value for the ObjectiveMechanism.  This
can be accomplished in several ways, that are illustrated in the examples below.  In the first example,
a `TransferMechanism` with the `SoftMax` function (and the `PROB <Softmax.PROB>` as its output format) to select
an action and represent its reward prediction.  This generates a vector with a single non-zero value, which designates
the predicted reward for the selected action.  Because the output is a vector, by default the InputPort of the
ObjectiveMechanism created to monitor it will also be a vector.  However, the ObjectiveMechanism requires this to be
a single value, that it can compare with the value of the reward Mechanism (monitoring the feedback provided by the
environment).  In the example below, this is accomplished by using `default_variable` in the constructor of the
ObjectiveMechanism to force the InputPort for the ObjectiveMechanism to have a single value::

    >>> import psyneulink as pnl
    >>> my_action_select_mech = pnl.TransferMechanism(default_variable=[0, 0, 0],
    ...                                               function=pnl.SoftMax(output=pnl.PROB),
    ...                                               name='Action Selection Mech')

    >>> my_reward_mech = pnl.TransferMechanism(default_variable=[0],
    ...                                        name='Reward Mech')

    >>> my_objective_mech = pnl.ObjectiveMechanism(default_variable=[[0],[0]],
    ...                                            monitor=[my_action_select_mech, my_reward_mech])

Note that the OutputPorts for the ``my_action_selection`` and ``my_reward_mech`` are specified
in `monitor`.  If that were the only specification, the InputPort created for ``my_action_select_mech``
would be a vector of length 3.  This is overridden by specifying `default_variable` as an array with two
single-value arrays (one corresponding to ``my_action_select_mech`` and the other to ``my_reward_mech``).  This forces
the InputPort for ``my_action_select_mech`` to have only a single element which, in turn, will cause a
MappingProjection to be created from  ``my_action_select_mech`` to the ObjectiveMechanism's InputPort using a
`FULL_CONNECTIVITY_MATRIX` (the one used for `AUTO_ASSIGN_MATRIX` when the sender and receiver have values of
different lengths).  This produces the desired effect, since the action selected is the only non-zero value in the
output of ``my_action_select_mech``, and so the `FULL_CONNECTIVITY_MATRIX` will combine it with zeros (the other values
in the vector), and so its value will be assigned as the value of the corresponding InputPort in the
ObjectiveMechanism.

An alternative would be to explicitly specify the `variable <InputPort.variable>` attribute for the InputPort created
for ``my_action_select_mech`` using a `InputPort specification dictionary <InputPort_Specification_Dictionary>` in
the **monitor** argument of ``my_objective_mech``, as follows::

    >>> my_objective_mech = pnl.ObjectiveMechanism(monitor=[{pnl.MECHANISM: my_action_select_mech,
    ...                                                      pnl.VARIABLE: [0]},
    ...                                                      my_reward_mech])

Note that the *VARIABLE* entry here specifies the `variable <InputPort.variable>` for the InputPort of the
ObjectiveMechanism created to receive a Projection from ``my_action_select_mech``, and not ``my_action_select_mech``
itself (see `ObjectiveMechanism_Input` for a full explanation).

.. _ObjectiveMechanism_Projection_Example:

Another way to specify the `variable <InputPort.variable>` for the InputPort of an ObjectiveMechanism is to
specify the Projections it receives from the OutputPort it monitors.  The following example uses a `tuple
specification <InputPort_Tuple_Specification:>` to assign the matrix for the MappingProjection from
``my_action_select_mech`` to the corresponding InputPort of ``my_objective_mech``::

    >>> import numpy as np
    >>> my_objective_mech = pnl.ObjectiveMechanism(monitor=[(my_action_select_mech, np.ones((3,1))), my_reward_mech])

Since the matrix specified has three rows (for its inputs) and one col (for the output), it will take the length three
vector provided as the output of ``my_action_select_mech`` and combine its elements into a single value that is
provided to the InputPort of the ObjectiveMechanism.

A `Connection tuple <InputPort_Tuple_Specification>` could also be used to specify the matrix, but this would require
that additional entries (for the weight and exponent of the InputPort) which, in this case, are not necessary (see
`example <ObjectiveMechanism_Tuple_Specification_Example>` below for how these can be used).

.. _ObjectiveMechanism_Weights_and_Exponents_Example:

By default, an ObjectiveMechanism simply adds the values received by each of its InputPorts to generate its output.
However, this too can be customized in a variety of ways.

.. _ObjectiveMechanism_Tuple_Specification_Example:

The simplest way is to assign values to the `weight <InputPort.weight>` and/or `exponent <InputPort.exponent>`
attributes of its InputPorts.  This can be done by placing them in a `tuple specification
<InputPort_Tuple_Specification>` for the OutputPort that provides value for the InputPort. In the example
below, the ObjectiveMechanism used in the previous example is further customized to subtract the value of the action
selected from the value of the reward::

    >>> my_objective_mech = pnl.ObjectiveMechanism(default_variable = [[0],[0]],
    ...                                            monitor = [(my_action_select_mech, -1, 1), my_reward_mech])

This specifies that ``my_action_select_mech`` should be assigned a weight of -1 and an exponent of 1 when it is
submitted to the ObjectiveMechanism's `function <ObjectiveMechanism.function>`.  Notice that the exponent had to be
included, even though it is the default value;  when a tuple is used, the weight and exponent values must both be
specified.  Notice also that ``my_reward_mech`` does not use a tuple, so it will be assigned defaults for both the
weight and exponent parameters.

.. _ObjectiveMechanism_Multiple_OutputPorts_Example:

An ObjectiveMechanism can also be configured to monitor multiple OutputPorts of the same Mechanism.  In the following
example, an ObjectiveMechanism is configured to calculate the reward rate for a `DDM` Mechanism, by specifying
OutputPorts for the DDM that report its response time and accuracy::

    >>> my_decision_mech = pnl.DDM(output_ports=[pnl.RESPONSE_TIME,
    ...                                           pnl.PROBABILITY_UPPER_THRESHOLD])

    >>> my_objective_mech = pnl.ObjectiveMechanism(monitor=[
    ...                                              my_reward_mech,
    ...                                              my_decision_mech.output_ports[pnl.PROBABILITY_UPPER_THRESHOLD],
    ...                                              (my_decision_mech.output_ports[pnl.RESPONSE_TIME], 1, -1)])

This specifies that the ObjectiveMechanism should multiply the `value <OutputPort.value>` of ``my_reward_mech``'s
`primary OutputPort <OutputPort_Primary>` by the `value <OutpuState.value>` of ``my_decision_mech``'s
*PROBABILITY_UPPER_THRESHOLD*, and divide the result by ``my_decision_mech``'s *RESPONSE_TIME* `value
<OutputPort.value>`.  The two OutputPorts of ``my_decision_mech`` are referenced as items in the `output_ports
<Mechanism_Base.output_ports>` list of ``my_decision_mech``.  However, a `2-item (Port name, Mechanism) tuple
<InputPort_Port_Mechanism_Tuple>` can be used to reference them more simply, as follows::

    >>> my_objective_mech = pnl.ObjectiveMechanism(monitor=[
    ...                                           my_reward_mech,
    ...                                           (pnl.PROBABILITY_UPPER_THRESHOLD, my_decision_mech),
    ...                                           ((pnl.RESPONSE_TIME, my_decision_mech), 1, -1)])


*Customizing the ObjectiveMechanism's function*

In the examples above, the weights and exponents assigned to the InputPorts are passed to the ObjectiveMechanism's
`function <ObjectiveMechanism.function>` for use in combining their values.  The same can be accomplished by
specifying the relevant parameter(s) of the function itself, as in the following example::

    >>> my_objective_mech = pnl.ObjectiveMechanism(default_variable = [[0],[0]],
    ...                                            monitor = [my_action_select_mech, my_reward_mech],
    ...                                            function=pnl.LinearCombination(weights=[[-1], [1]]))

Here, the `weights <LinearCombination.weights>` parameter of the `LinearCombination` function is specified directly,
with two values [-1] and [1] corresponding to the two items in `monitor` (and `default_variable`).
This will multiply the value from ``my_action_select_mech`` by -1 before adding it to (and thus subtracting it from)
the value of ``my_reward_mech``.  Notice that the weight for ``my_reward_mech`` had to be specified, even though it
is using the default value (1);  whenever a weight and/or exponent parameter is specified, there must be an entry for
every item of the function's variable.  However, the `exponents <LinearCombination.exponents>` did not need to be
specified, as it is not used.  However it, and the `operation <LinearCombination.operation>` parameters of
`LinearCombination` can also be used to multiply and divide quantities.

COMMENT:
**ADD DISCUSSION OF Projection WEIGHTS AND EXPONENTS ONCE THEY ARE IMPLEMENTED FOR HIERARCHICAL OPERATIONS
COMMENT

.. _ObjectiveMechanism_Class_Reference:

Class Reference
---------------

"""

import warnings
from collections import namedtuple
from collections.abc import Iterable

import numpy as np
from beartype import beartype

from psyneulink._typing import Optional, Union

from psyneulink.core.components.functions.nonstateful.transformfunctions import LinearCombination
from psyneulink.core.components.mechanisms.mechanism import MechanismError
from psyneulink.core.components.mechanisms.processing.processingmechanism import ProcessingMechanism_Base
from psyneulink.core.components.ports.inputport import InputPort, INPUT_PORT
from psyneulink.core.components.ports.outputport import OutputPort
from psyneulink.core.components.ports.port import _parse_port_spec
from psyneulink.core.globals.context import Context, ContextFlags, handle_external_context
from psyneulink.core.globals.keywords import \
    CONTROL, EXPONENT, EXPONENTS, LEARNING, MATRIX, NAME, OBJECTIVE_MECHANISM, OUTCOME, OWNER_VALUE, \
    PARAMS, PREFERENCE_SET_NAME, PROJECTION, PROJECTIONS, PORT_TYPE, VARIABLE, WEIGHT, WEIGHTS
from psyneulink.core.globals.parameters import Parameter, check_user_specified
from psyneulink.core.globals.preferences.basepreferenceset import ValidPrefSet, REPORT_OUTPUT_PREF
from psyneulink.core.globals.preferences.preferenceset import PreferenceEntry, PreferenceLevel
from psyneulink.core.globals.utilities import ContentAddressableList

__all__ = [
    'DEFAULT_MONITORED_PORT_WEIGHT', 'DEFAULT_MONITORED_PORT_EXPONENT', 'DEFAULT_MONITORED_PORT_MATRIX',
    'MONITOR', 'MONITOR_SUFFIX', 'MONITORED_OUTPUT_PORT_NAME_SUFFIX', 'MONITORED_OUTPUT_PORTS',
    'ObjectiveMechanism', 'ObjectiveMechanismError', 'OUTCOME', 'ROLE'
]

ROLE = 'role'
MONITOR = 'monitor'
MONITOR_SUFFIX = '_Monitor'
MONITORED_OUTPUT_PORTS = 'monitored_output_ports'
MONITORED_OUTPUT_PORT_NAME_SUFFIX = '_Monitor'

DEFAULT_MONITORED_PORT_WEIGHT = None
DEFAULT_MONITORED_PORT_EXPONENT = None
DEFAULT_MONITORED_PORT_MATRIX = None


class ObjectiveMechanismError(MechanismError):
    pass


class ObjectiveMechanism(ProcessingMechanism_Base):
    """
    ObjectiveMechanism(             \
        monitor,                    \
        function=LinearCombination, \
        output_ports=OUTCOME)

    Subclass of `ProcessingMechanism <ProcessingMechanism>` that evaluates the value(s) of one or more `OutputPorts
    <OutputPort>`.  See `Mechanism <Mechanism_Class_Reference>` for additional arguments and attributes.


    Arguments
    ---------

    monitor : List[`OutputPort`, 'InputPort`, `Mechanism`, str, value, dict, `MonitoredOutputPortsOption`] or dict
        specifies the OutputPorts, the `values <OutputPort.value>` of which will be monitored, and evaluated by
        `function <ObjectiveMechanism.function>` (see `ObjectiveMechanism_Monitor` for details of specification).

    function : TransformFunction, ObjectiveFunction, function or method : default LinearCombination
        specifies the function used to evaluate the values listed in `monitor` <ObjectiveMechanism.monitor>`
        (see `function <ObjectiveMechanism.function>` for details).

    output_ports :  list[OutputPort, value, str or dict] or dict[] : default [OUTCOME]
        specifies the OutputPorts for the Mechanism;

    COMMENT:
    role: LEARNING or CONTROL : default None
        specifies if the ObjectiveMechanism is being used for learning or control (see `role` for details).
    COMMENT

    Attributes
    ----------

    monitor : ContentAddressableList[OutputPort]
        determines the OutputPorts, the `values <OutputPort.value>` of which are monitored, and evaluated by the
        ObjectiveMechanism's `function <ObjectiveMechanism.function>`.  Each item in the list refers to an
        `OutputPort` containing the value to be monitored, with a `MappingProjection` from it to an
        corresponding `InputPort` listed in the `input_ports <ObjectiveMechanism.input_ports>` attribute

        .. note::
           If any of the ObjectiveMechanism's `input_ports <ObjectiveMechanism.input_ports>` were specified to `shadow
           the InputPort <InputPort_Shadow_Inputs>` of another Mechanism, and any of those shadowed InputPorts
           receives more than one `Projection`, then the list of monitored OutputPorts in `monitor` will be longer
           than the list of the ObjectiveMechanism's `input_ports <ObjectiveMechanisms.input_ports>`.

    monitor_weights_and_exponents : List[Tuple(float, float)]
        each tuple in the list contains a weight and exponent associated with a corresponding InputPort listed in the
        ObjectiveMechanism's `input_ports <ObjectiveMechanism.input_ports>` attribute;  these are used by its
        `function <ObjectiveMechanism.function>` to parametrize the contribution that the values of each of the
        OuputStates monitored by the ObjectiveMechanism makes to its output (see `ObjectiveMechanism_Function`)

    input_ports : ContentAddressableList[InputPort]
        contains the InputPorts of the ObjectiveMechanism, each of which receives a `MappingProjection` from the
        OutputPorts specified in its `monitor <ObjectiveMechanism.monitor>` attribute.

    function : TransformFunction, ObjectiveFunction, function, or method
        the function used to evaluate the values monitored by the ObjectiveMechanism.  The function can be any
        `TransformFunction <Transformfunctions>` or a Python function that takes a 2d array with an arbitrary
        number of items or a number equal to the number of items in the ObjectiveMechanism's variable (i.e.,
        its number of input_ports) and returns a 1d array.

    output_port : OutputPort
        contains the `primary OutputPort <OutputPort_Primary>` of the ObjectiveMechanism; the default is
        its *OUTCOME* `OutputPort <ObjectiveMechanism_Output>`, the value of which is equal to the
        `value <Mechanism_Base.value>` attribute of the ObjectiveMechanism.

    output_ports : ContentAddressableList[OutputPort]
        by default, contains only the *OUTCOME* (`primary <OutputPort_Primary>`) OutputPort of the ObjectiveMechanism.

    output_values : 2d np.array
        contains one item that is the value of the *OUTCOME* `OutputPort <ObjectiveMechanism_Output>`.

    standard_output_ports : list[str]
        list of `Standard OutputPort <OutputPort_Standard>` that includes the following in addition to
        the `standard_output_ports <Mechanism_Base.standard_output_ports>` of a `Mechanism <Mechanism>`:

        .. _OBJECTIVE_MECHANISM_OUTCOME

        *OUTCOME* : 1d np.array
            the value of the objective or "loss" function computed by the
            ObjectiveMechanism's `function <ObjectiveMechanism.function>`

    modulatory_mechanism : None or ModulatoryMechanism
        `ModulatoryMechanism` to which ObjectiveMechanism has been assigned, and to which one or more of its
        `output_ports <ObjectiveMechanism.output_ports>` project. If that is a `ControlMechanism`, then
        the ObjectiveMechanism is assigned as its `objective_mechanism <ControlMechanism.objective_mechanism>`;
        if it is a `LearningMechanism`, it is in its `error_sources <LearningMechanism.error_sources>` attribute.

    """

    componentType = 'ObjectiveMechanism'

    classPreferenceLevel = PreferenceLevel.SUBTYPE
    # These will override those specified in TYPE_DEFAULT_PREFERENCES
    classPreferences = {
        PREFERENCE_SET_NAME: 'ObjectiveCustomClassPreferences',
        REPORT_OUTPUT_PREF: PreferenceEntry(False, PreferenceLevel.INSTANCE)}

    # class_defaults.variable = None;  Must be specified using either **input_ports** or **monitor**
    # kmantel: above needs to be clarified - can class_defaults.variable truly be anything? or should there be some format?
    #   if the latter, we should specify one such valid assignment here, and override _validate_default_variable accordingly
    class Parameters(ProcessingMechanism_Base.Parameters):
        """
            Attributes
            ----------

                function
                    see `function <ObjectiveMechanism_Function>`

                    :default value: `LinearCombination`
                    :type: `Function`

                input_ports
                    see `input_ports <ObjectiveMechanism.input_ports>`

                    :default value: None
                    :type:
                    :read only: True

                output_ports
                    see `output_ports <ObjectiveMechanism.output_ports>`

                    :default value: [`OUTCOME`]
                    :type: ``list``
                    :read only: True
        """
        function = Parameter(LinearCombination, stateful=False, loggable=False)

        input_ports = Parameter(
            None,
            stateful=False,
            loggable=False,
            read_only=True,
            structural=True,
            parse_spec=True,
            aliases='monitor',
            constructor_argument='monitor'
        )
        output_ports = Parameter(
            [OUTCOME],
            stateful=False,
            loggable=False,
            read_only=True,
            structural=True,
        )

    # ObjectiveMechanism parameter and control signal assignments):

    standard_output_ports = ProcessingMechanism_Base.standard_output_ports.copy()
    standard_output_ports.extend([{NAME:OUTCOME, VARIABLE:(OWNER_VALUE, 0)}])
    standard_output_port_names = ProcessingMechanism_Base.standard_output_ports.copy()
    standard_output_port_names.extend([OUTCOME])

    # FIX:  TYPECHECK MONITOR TO LIST OR ZIP OBJECT
    @check_user_specified
    @beartype
    def __init__(self,
                 monitor=None,
                 default_variable=None,
                 input_shapes=None,
                 function=None,
                 output_ports: Optional[Union[str, Iterable]] = None,
                 params=None,
                 name=None,
                 prefs: Optional[ValidPrefSet] = None,
                 **kwargs):

        # For backward compatibility
        if MONITORED_OUTPUT_PORTS in kwargs:
            if monitor:
                raise ObjectiveMechanismError(f'Can\'t specifiy both {repr(MONITOR)} ({monitor}) '
                                              f'and {repr(MONITORED_OUTPUT_PORTS)} ({kwargs[MONITORED_OUTPUT_PORTS]})'
                                              f' args of {self.name} specified; pick one!')
            warnings.warn(f'Use of {repr(MONITORED_OUTPUT_PORTS)} as arg of {self.__class__.__name__} is deprecated;  '
                          f'use {repr(MONITOR)} instead')
            monitor = kwargs.pop(MONITORED_OUTPUT_PORTS)
        monitor = monitor or None # deal with possibility of empty list
        if output_ports is None or output_ports == OUTCOME:
            output_ports = [OUTCOME]

        super().__init__(
            default_variable=default_variable,
            input_shapes=input_shapes,
                         monitor=monitor,
                         output_ports=output_ports,
                         function=function,
                         params=params,
                         name=name,
                         prefs=prefs,
                         **kwargs)

        # This is used to specify whether the ObjectiveMechanism is associated with a ControlMechanism that is
        #    the controller for a Composition;  it is set by the ControlMechanism when it creates the ObjectiveMechanism
        self.for_controller = False
        self.modulatory_mechanism = None

    def _validate_params(self, request_set, target_set=None, context=None):
        """Validate **role**, **monitor**, amd **input_ports** arguments

        """

        super()._validate_params(request_set=request_set,
                                 target_set=target_set,
                                 context=context)

        if ROLE in target_set and target_set[ROLE] and not target_set[ROLE] in {LEARNING, CONTROL}:
            raise ObjectiveMechanismError("\'role\'arg ({}) of {} must be either \'LEARNING\' or \'CONTROL\'".
                                          format(target_set[ROLE], self.name))

    def _instantiate_input_ports(self, monitor_specs=None, reference_value=None, context=None):
        """Instantiate InputPorts specified in **input_ports** argument of constructor or each OutputPort
        specified in monitor_specs

        Called during initialization as well as by _add_to_monitor(),
            so must distinguish between initialization and adding to instantiated input_ports.

        During initialization, uses **input_ports** as specification of InputPorts to instantiate;
            if none are specified, instantiates a default InputPort

        Otherwise, uses monitor_specs as specification of InputPorts to instantiate;
            these will replace any existing InputPorts (including a default one)
        """

        # If call is for initialization
        if self.initialization_status == ContextFlags.INITIALIZING:
            # Use self.input_ports (containing specs from **input_ports** arg of constructor)
            #    or pass off instantiation of default InputPort(s) to super
            input_ports = self.input_ports or None
            input_ports = super()._instantiate_input_ports(input_ports=input_ports, context=context)
        else:
            # Parse any spec that is an InputPort into the OutputPort(s) that project to it
            monitor_specs = _parse_monitor_specs(monitor_specs)

            # Instantiate InputPorts corresponding to items specified in monitor
            #     (note: these will replace any existing ones, including the default one created on initialization)
            input_ports = super()._instantiate_input_ports(input_ports=monitor_specs,
                                                             reference_value=reference_value,
                                                             context=context)
        self._name_input_ports(input_ports)
        return input_ports

    def _name_input_ports(self, input_ports):
        # If InputPorts are not already named, name them based on the OutputPorts that project to them
        from psyneulink.core.globals.registry import remove_instance_from_registry, register_instance
        if not input_ports:
            return
        for port in input_ports:
            if not port.path_afferents:
                continue
            if len(port.path_afferents) > 1:
                assert False
            # If the name is not a default name, return
            if not (port.name is InputPort.__name__ or InputPort.__name__ + '-' in port.name):
                return
            proj = port.path_afferents[0]
            remove_instance_from_registry(registry=self._portRegistry,
                                          category=INPUT_PORT,
                                          component=port)
            port.name = "Value of {} [{}]".format(proj.sender.owner.name, proj.sender.name)
            register_instance(port, port.name, InputPort, self._portRegistry, INPUT_PORT)

    @handle_external_context(source=ContextFlags.METHOD)
    def add_to_monitor(self, monitor_specs, context=None):
        """Instantiate `OutputPorts <OutputPort>` to be monitored by the ObjectiveMechanism.

        Used by other Components to add a `Port` or list of Ports to be monitored by the ObjectiveMechanism.
        The **monitor_spec** can be any of the following:
        - MonitoredOutputPortTuple
        - `Mechanism <Mechanism>`;
        - `OutputPort`;
        - `tuple specification <InputPort_Tuple_Specification>`;
        - `Port specification dictionary <InputPort_Specification_Dictionary>`;
        - list with any of the above.
        If the item is a Mechanism, its `primary OutputPort <OutputPort_Primary>` is used.
        """
        monitor_specs = list(monitor_specs)

        # If ObjectiveMechanism has only its default InputPort and that has no afferent Projections:
        #    delete it and first item of variable
        if len(self.input_ports)==1 and self.input_port.name=='InputPort-0' and not self.input_port.path_afferents:
            del self.input_ports[0]
            self.defaults.variable = []

        # Get reference value
        reference_value = []
        # Get value of each OutputPort or, if a Projection from it is specified, then the Projection's value
        for i, spec in enumerate(monitor_specs):
            from psyneulink.core.components.ports.inputport import InputPort
            from psyneulink.core.components.mechanisms.modulatory.control.controlmechanism import MonitoredOutputPortTuple
            from psyneulink.core.components.projections.projection import _get_projection_value_shape

            # If it is a MonitoredOutputPortTuple, create InputPort specification dictionary
            if isinstance(spec, MonitoredOutputPortTuple):
                # If matrix is specified, let it determine the variable
                if spec.matrix is not None:
                    variable = None
                # Otherwise, use OutputPort's value as variable for InputPort
                else:
                    variable = spec.output_port.value
                spec = {NAME: spec.output_port.name,
                        VARIABLE: variable,
                        WEIGHT: spec.weight,
                        EXPONENT: spec.exponent,
                        PROJECTIONS: [(spec.output_port, spec.matrix)]}
                monitor_specs[i] = spec

            # Parse spec to get value of OutputPort and (possibly) the Projection from it
            input_port = _parse_port_spec(owner=self, port_type = InputPort, port_spec=spec,
                                          context=Context(string='objective_mechanism.add_to_monitor'))

            # There should be only one ProjectionTuple specified,
            #    that designates the OutputPort and (possibly) a Projection from it
            if len(input_port[PARAMS][PROJECTIONS])!=1:
                raise ObjectiveMechanismError("PROGRAM ERROR: Failure to parse item in monitor_specs "
                                              "for {} (item: {})".format(self.name, spec))
            projection_tuple = input_port[PARAMS][PROJECTIONS][0]
            # If Projection is specified, use its value
            if PROJECTION in projection_tuple.projection:
                reference_value.append(projection_tuple.projection[PROJECTION].value)
            # If matrix is specified for Projection, get its receiver dimension
            elif MATRIX in projection_tuple.projection:
                reference_value.append(_get_projection_value_shape(projection_tuple.port,
                                                                   projection_tuple.projection[MATRIX]))
            # Otherwise, use its sender's (OutputPort) value
            else:
                reference_value.append(projection_tuple.port.value)

        input_ports = self._instantiate_input_ports(monitor_specs=monitor_specs,
                                                      reference_value=reference_value,
                                                      context=context
                                                      )

        output_ports = [[projection.sender for projection in port.path_afferents] for port in input_ports]

        self._instantiate_function_weights_and_exponents(context=context)

        return output_ports

    def _instantiate_attributes_after_function(self, context=None):
        """Assign InputPort weights and exponents to ObjectiveMechanism's function
        """
        super()._instantiate_attributes_after_function(context=context)
        self._instantiate_function_weights_and_exponents(context=context)

    def _instantiate_function_weights_and_exponents(self, context=None):
        """Assign weights and exponents to ObjectiveMechanism's function if it has those attributes

        For each, only make assignment if one or more entries in it has been assigned a value
        If any one value has been assigned, assign default value (1) to all other elements
        """
        DEFAULT_WEIGHT = 1
        DEFAULT_EXPONENT = 1

        weights = [input_port.defaults.weight for input_port in self.input_ports]
        exponents = [input_port.defaults.exponent for input_port in self.input_ports]

        if WEIGHTS in self.function.parameters:
            if any(weight is not None for weight in weights):
                self.function.parameters.weights._set(
                    np.asarray([
                        [weight if weight is not None else DEFAULT_WEIGHT]
                        for weight in weights
                    ]),
                    context
                )
        if EXPONENTS in self.function.parameters:
            if any(exponent is not None for exponent in exponents):
                self.function.parameters.exponents._set(
                    np.asarray([
                        [exponent if exponent is not None else DEFAULT_EXPONENT]
                        for exponent in exponents
                    ]),
                    context
                )
        assert True

    # # MODIFIED 6/8/19 NEW: [JDC]
    # def _parse_function_variable(self, variable, context=None, context=None):
    #     # CRASHES IN x_or TEST AND LLVM TESTS:
    #     if self.function.variableEncodingDim < self.variableEncodingDim:
    #         return np.squeeze(variable)
    #     # if self.function.variableEncodingDim > self.variableEncodingDim:
    #     #     return np.expand_dims(variable,0)
    #     return variable
    #
    #     # assert False
    # # MODIFIED 6/8/19 END

    @property
    def monitor(self):
        if not isinstance(self.input_ports, ContentAddressableList):
            return None
        else:
            monitor = []
            for input_port in self.input_ports:
                for projection in input_port.path_afferents:
                    monitor.append(projection.sender)

            return ContentAddressableList(component_type=OutputPort,
                                          list=[projection.sender for input_port in self.input_ports
                                                for projection in input_port.path_afferents])

    @property
    def monitor_weights_and_exponents(self):
        if hasattr(self.function, WEIGHTS) and self.function.weights.base is not None:
            weights = self.function.weights.base
        else:
            weights = [input_port.weight.base for input_port in self.input_ports]
        if hasattr(self.function, EXPONENTS) and self.function.exponents.base is not None:
            exponents = self.function.exponents.base
        else:
            exponents = [input_port.exponent.base for input_port in self.input_ports]
        return [(w,e) for w, e in zip(weights,exponents)]

    @monitor_weights_and_exponents.setter
    def monitor_weights_and_exponents(self, weights_and_exponents_tuples):

        weights = [w[0] for w in weights_and_exponents_tuples]
        exponents = [e[1] for e in weights_and_exponents_tuples]
        self._instantiate_weights_and_exponents(weights, exponents)

    # For backward compatibility
    @property
    def monitored_output_ports(self):
        return self.monitor

    # For backward compatibility
    @property
    def monitored_output_ports_weights_and_exponents(self):
        return self.monitor_weights_and_exponents

    # For backward compatibility
    @monitored_output_ports_weights_and_exponents.setter
    def monitored_output_ports_weights_and_exponents(self, weights_and_exponents_tuples):
        self.monitor_weights_and_exponents = weights_and_exponents_tuples

# FIX: 11/3/21 -- MOVE THIS TO ControlMechanism, AND INTEGRATE WITH ControlMechanism.validate_monitored_port_spec()
#                 OR MOVE THAT METHOD TO HERE??
def _parse_monitor_specs(monitor_specs):
    spec_tuple = namedtuple('SpecTuple', 'index spec')
    parsed_specs = []
    specs_to_replace = []
    for i, spec in enumerate(parsed_specs):
        if (isinstance(spec, InputPort)
                or (isinstance(spec, dict) and spec[PORT_TYPE] in {INPUT_PORT, InputPort})):
            pass # DO PARSING HERE
            if isinstance(spec, InputPort):
                pass # Get all projects to it and create InputPort specification dictionary with them
            if isinstance(spec, dict) and spec[PORT_TYPE] in {INPUT_PORT, InputPort}:
                pass # Get Projections entry and append
            new_spec = None
            specs_to_replace.append(spec_tuple(i, new_spec))

    for spec_tuple in specs_to_replace:
        # # Delete old spec in monitored_specs at specified index
        # del monitor_specs[spec_tuple.index]
        # # Insert new item(s) from spec_tuple into monitor_specs at specified index
        # monitor_specs[spec_tuple.index:spec_tuple.index] = spec_tuple.spec
        monitor_specs[spec_tuple.index] = spec_tuple.spec
    return monitor_specs

# IMPLEMENTATION NOTE:  THIS SHOULD BE MOVED TO COMPOSITION ONCE THAT IS IMPLEMENTED
#                      ??MAYBE INTEGRATE INTO Port MODULE (IN _instantate_port)
# KAM commented out _instantiate_monitoring_projections 9/28/18 to avoid confusion because it never gets called
# @beartype
# def _instantiate_monitoring_projections(
#     owner,
#     sender_list: Union[list, ContentAddressableList],
#     receiver_list: Union[list, ContentAddressableList],
#     receiver_projection_specs: Optional[list]=None,
#     system=None,
#     context=None
# ):
#
#     from psyneulink.core.components.ports.outputPort import OutputPort
#     from psyneulink.core.components.projections.pathway.mappingprojection import MappingProjection
#     from psyneulink.core.components.projections.projection import ProjectionTuple
#
#     receiver_projection_specs = receiver_projection_specs or [DEFAULT_MATRIX] * len(sender_list)
#
#     if len(sender_list) != len(receiver_list):
#         raise ObjectiveMechanismError("PROGRAM ERROR: Number of senders ({}) does not equal number of receivers ({}) "
#                                      "in call to instantiate monitoring projections for {}".
#                                      format(len(sender_list), len(receiver_list), owner.name))
#
#     if len(receiver_projection_specs) != len(receiver_list):
#         raise ObjectiveMechanismError("PROGRAM ERROR: Number of projection specs ({}) "
#                                      "does not equal number of receivers ({}) "
#                                      "in call to instantiate monitoring projections for {}".
#                                      format(len(receiver_projection_specs), len(receiver_list), owner.name))
#
#     # Instantiate InputPort with Projection from OutputPort specified by sender
#     for sender, receiver, recvr_projs in zip(sender_list, receiver_list, receiver_projection_specs):
#         # IMPLEMENTATION NOTE:  If there is more than one Projection specified for a receiver, only the 1st is used;
#         #                           (there should only be one if a 2-item tuple was used to specify the InputPort,
#         #                            however other forms of specifications could produce more)
#         if isinstance(recvr_projs,list) and len(recvr_projs) > 1 and owner.verbosePref:
#             warnings.warn("{} projections were specified for InputPort ({}) of {} ;"
#                           "only the first ({}) will be used".
#                           format(len(recvr_projs), receiver.name, owner.name, recvr_projs[0].port.name))
#             projection_spec = recvr_projs[0]
#         else:
#             projection_spec = recvr_projs
#
#         if isinstance(projection_spec, ProjectionTuple):
#             projection_spec = projection_spec.projection
#
#         # IMPLEMENTATION NOTE:  This may not handle situations properly in which the OutputPort is specified
#         #                           by a 2-item tuple (i.e., with a Projection specification as its second item)
#         if isinstance(sender, OutputPort):
#             # Projection has been specified for receiver and initialization begun, so call deferred_init()
#             if receiver.path_afferents:
#                 if not receiver.path_afferents[0].initialization_status == ContextFlags.DEFERRED_INIT:
#                     raise ObjectiveMechanismError("PROGRAM ERROR: {} of {} already has an afferent projection "
#                                                   "implemented and initialized ({})".
#                                                   format(receiver.name, owner.name, receiver.path_afferents[0].name))
#                 # FIX: 10/3/17 - IS IT OK TO IGNORE projection_spec IF IT IS None?  SHOULD IT HAVE BEEN SPECIFIED??
#                 # FIX:           IN DEVEL, projection_spec HAS BEEN PROPERLY ASSIGNED
#                 if (projection_spec and
#                         not receiver.path_afferents[0].function_params[MATRIX] is projection_spec):
#                     raise ObjectiveMechanismError("PROGRAM ERROR: Projection specification for {} of {} ({}) "
#                                                   "does not match matrix already assigned ({})".
#                                                   format(receiver.name,
#                                                          owner.name,
#                                                          projection_spec,
#                                                          receiver.path_afferents[0].function_params[MATRIX]))
#                 receiver.path_afferents[0]._init_args[SENDER] = sender
#                 receiver.path_afferents[0]._deferred_init(context=context)
#             else:
#                 projection_spec = MappingProjection(sender=sender,
#                                                     receiver=receiver,
#                                                     matrix=projection_spec,
#                                                     name=sender.name + ' monitor')
#                 projection_spec._activate_for_compositions(system)



import inspect
import psyneulink as pnl
import numpy as np
import pytest
import re

from psyneulink.core.components.projections.modulatory.modulatoryprojection import ModulatoryProjection_Base
from psyneulink.core.components.projections.pathway.pathwayprojection import PathwayProjection_Base

# gather all Component classes (a set to ensure no duplicates)
component_classes = []
component_class_constructor_arguments = {}
for item in pnl.__all__:
    evaled = eval(f'pnl.{item}')

    if isinstance(
        evaled,
        pnl.core.components.component.ComponentsMeta
    ):
        component_classes.append(evaled)
        component_class_constructor_arguments[evaled] = inspect.signature(
            evaled.__init__
        ).parameters

component_classes.sort(key=lambda x: x.__name__)


@pytest.mark.parametrize(
    'class_type',
    [
        pnl.Mechanism_Base,
        pnl.Function_Base,
        pnl.Port_Base,
        ModulatoryProjection_Base,
        PathwayProjection_Base,
    ]
)
def test_abstract_classes(class_type):
    with pytest.raises(TypeError) as err:
        class_type()
    assert 'abstract class' in str(err.value)


@pytest.mark.parametrize(
    'class_',
    component_classes
)
def test_function_parameters_stateless(class_):
    try:
        assert class_.parameters.function.stateful is False, (
            f'{class_.__name__}.parameters.function.stateful is True. '
            'The function Parameter for Components is currently '
            'expected to be stateless (defined as stateful=False)'
        )
    except AttributeError:
        pass


@pytest.mark.parametrize("class_", component_classes)
def test_constructors_have_check_user_specified(class_):
    assert "check_user_specified" in inspect.getsource(class_.__init__), (
        f"The __init__ method of Component {class_.__name__} must be wrapped by"
        f" check_user_specified in {pnl.core.globals.parameters.check_user_specified.__module__}"
    )


def _numeric_parameter_value_check(class_, param_name, value, descriptor):
    descriptor = f'{class_}.parameters.{param_name}{descriptor}'
    assert isinstance(value, np.ndarray) or not pnl.is_numeric(value), (
        f'{descriptor} is a numeric value but is not wrapped in a'
        f' numpy array:\n\t{value}\n\t{type(value)}'
    )


# could parametrize over each parameter for each class instead of
# looping, but this would greatly increase the overall number of tests
# for minimal benefit
@pytest.mark.parametrize('class_', component_classes)
def test_numeric_parameter_values_are_numpy_defaults(class_):
    for parameter in class_.parameters:
        _numeric_parameter_value_check(
            class_, parameter.name, parameter.default_value, ' default_value'
        )


@pytest.mark.parametrize('class_', component_classes)
def test_numeric_parameter_values_are_numpy_values(class_):
    for parameter in class_.parameters:
        for eid, v in parameter.values.items():
            _numeric_parameter_value_check(class_, parameter.name, v, f'.values[{eid}]')


@pytest.fixture(scope='module')
def nested_compositions():
    comp = pnl.Composition(name='comp')
    inner_comp = pnl.Composition(name='Inner Composition')
    A = pnl.TransferMechanism(
        function=pnl.Linear(slope=5.0, intercept=2.0),
        name='A'
    )
    B = pnl.TransferMechanism(function=pnl.Logistic, name='B')
    C = pnl.RecurrentTransferMechanism(name='C')
    D = pnl.IntegratorMechanism(
        function=pnl.SimpleIntegrator(noise=pnl.NormalDist()),
        name='D'
    )
    E = pnl.TransferMechanism(name='E')
    F = pnl.TransferMechanism(name='F')

    for m in [E, F]:
        inner_comp.add_node(m)

    for m in [A, B, C, D, inner_comp]:
        comp.add_node(m)

    comp.add_projection(pnl.MappingProjection(), A, B)
    comp.add_projection(pnl.MappingProjection(), A, C)
    comp.add_projection(pnl.MappingProjection(), B, D)
    comp.add_projection(pnl.MappingProjection(), C, D)
    comp.add_projection(pnl.MappingProjection(), C, inner_comp)

    inner_comp.add_projection(pnl.MappingProjection(), E, F)

    yield comp, inner_comp


@pytest.mark.parametrize(
    'filter_name, filter_regex, unknown_param_names',
    [
        (None, None, []),
        (None, 'slo$', ['slope']),
        ('slo', None, ['slope']),
        ('slo', 'slo$', ['slope']),
        (['slope', 'seed'], None, []),
        (None, ['slope', 'seed'], []),
        (None, ['.*_param'], ['slope']),
    ]
)
def test_all_dependent_parameters(
    nested_compositions,
    filter_name,
    filter_regex,
    unknown_param_names
):
    comp, inner_comp = nested_compositions

    params_comp = comp.all_dependent_parameters(filter_name, filter_regex)
    params_inner_comp = inner_comp.all_dependent_parameters(
        filter_name, filter_regex
    )

    params_comp_keys = set(params_comp.keys())
    params_inner_comp_keys = set(params_inner_comp.keys())

    assert params_inner_comp_keys.issubset(params_comp_keys)
    assert (
        len(params_comp_keys) == 0
        or not params_comp_keys.issubset(params_inner_comp_keys)
    )

    if filter_name is not None:
        if isinstance(filter_name, str):
            filter_name = [filter_name]

    if filter_regex is not None:
        if isinstance(filter_regex, str):
            filter_regex = [filter_regex]

    for item, comp_name in [
        (params_comp, 'comp'),
        (params_inner_comp, 'inner_comp')
    ]:
        for p in item:
            assert p._owner._owner is item[p], (p.name, comp_name)

            matches = True
            try:
                matches = matches and p.name in filter_name
            except TypeError:
                pass

            try:
                for pattern in filter_regex:
                    matches = matches or re.match(pattern, p.name)
            except TypeError:
                pass

            assert matches

        for p in unknown_param_names:
            assert p not in item, (p.name, comp_name)



import copy

import optuna.samplers
from fastkde import fastKDE
from scipy.interpolate import interpn
from scipy.optimize import differential_evolution
from beartype import beartype

from psyneulink.core.globals import SampleIterator
from psyneulink.core.globals.context import Context, ContextFlags, handle_external_context
from psyneulink.core.components.functions.nonstateful.optimizationfunctions import (
    OptimizationFunction,
    OptimizationFunctionError,
    SEARCH_SPACE,
)
from psyneulink.core.globals.parameters import SharedParameter, check_user_specified
from psyneulink.core.globals.utilities import try_extract_0d_array_item

from psyneulink._typing import (
    Dict,
    Tuple,
    Callable,
    List,
    Optional,
    Union,
    Type,
    Literal,
    Mapping,
)


import time
import numpy as np

from rich.progress import Progress, BarColumn, TimeRemainingColumn

import warnings
import logging

from rich.markup import escape

logger = logging.getLogger(__name__)

__all__ = ["PECOptimizationFunction", "BadLikelihoodWarning", "PECObjectiveFuncWarning"]


def get_param_str(params):
    """
    A simple function to turn a dict into a string with commas separating key=value pairs.

    Parameters
    ----------
    params: The parameters to print.

    Returns
    -------
    The string version of the parameter dict

    """
    return ", ".join(
        f"[dodger_blue1]{escape(name.replace('PARAMETER_CIM_', ''))}[/dodger_blue1]=[spring_green1]{value:.5f}[/spring_green1]"
        for name, value in params.items()
    )


class PECObjectiveFuncWarning(UserWarning):
    """
    A custom warning that is used to signal when the objective function could not be evaluated for some reason.
    This is usually caused when parameter values cause degenerate simulation results (no variance in values).
    """

    pass


class BadLikelihoodWarning(PECObjectiveFuncWarning):
    """
    A custom warning that is used to signal when the likelihood could not be evaluated for some reason.
    This is usually caused when parameter values cause degenerate simulation results (no variance in values).
    It can also be caused when experimental data is not matching any of the simulation results.
    """

    pass


def simulation_likelihood(
    sim_data, exp_data=None, categorical_dims=None, combine_trials=False
):
    """
    Compute the likelihood of a simulation dataset (or the parameters that generated it) conditional
    on a set of experimental data. This function essentially just computes the kernel density estimate (KDE)
    of the simulation data at the experimental data points. If no experimental data is provided just return
    the KDE evaluated at default points provided by the fastkde library.

    Some related work:

    Steven Miletić, Brandon M. Turner, Birte U. Forstmann, Leendert van Maanen,
    Parameter recovery for the Leaky Competing Accumulator model,
    Journal of Mathematical Psychology,
    Volume 76, Part A,
    2017,
    Pages 25-50,
    ISSN 0022-2496,
    https://doi.org/10.1016/j.jmp.2016.12.001.
    (http://www.sciencedirect.com/science/article/pii/S0022249616301663)

    This function uses the wonderful fastKDE package:

    O’Brien, T. A., Kashinath, K., Cavanaugh, N. R., Collins, W. D. & O’Brien, J. P.
    A fast and objective multidimensional kernel density estimation method: fastKDE.
    Comput. Stat. Data Anal. 101, 148–160 (2016).
    <http://dx.doi.org/10.1016/j.csda.2016.02.014>__

    O’Brien, T. A., Collins, W. D., Rauscher, S. A. & Ringler, T. D.
    Reducing the computational cost of the ECF using a nuFFT: A fast and objective probability density estimation method.
    Comput. Stat. Data Anal. 79, 222–234 (2014). <http://dx.doi.org/10.1016/j.csda.2014.06.002>__

    Parameters
    ----------
    sim_data: Data collected over many simulations. This must be either a 2d or 3d numpy array.
        If 2D, the first dimension is the simulation number and the second dimension is data points. That is,
        each row is a simulation. If 3d, the first dimension is the trial, the second dimension is the
        simulation number, and the final dimension is data points.

    exp_data: This must be a numpy array with identical format as the simulation data, with the exception
        that there is no simulation dimension.

    categorical_dims: a list of indices that indicate categorical dimensions of a data point. Length must be
        the same length as last dimension of sim_data and exp_data.

    combine_trials: Combine data across all trials into a single likelihood estimate, this assumes
        that the parameters of the simulations are identical across trials.

    Returns
    -------
    The pdf of simulation data (or in other words, the generating parameters) conditioned on the
    experimental data.

    """

    # Add a singleton dimension for trials if needed.
    if sim_data.ndim == 2:
        sim_data = sim_data[None, :, :]

    if combine_trials and sim_data.shape[0] > 1:
        sim_data = np.vstack(sim_data)[None, :, :]

    if type(categorical_dims) != np.ndarray:
        categorical_dims = np.array(categorical_dims)

    con_sim_data = sim_data[:, :, ~categorical_dims]
    cat_sim_data = sim_data[:, :, categorical_dims]

    categories = np.unique(cat_sim_data)

    if len(categories) > 10:
        raise ValueError("Too many unique values present for a categorical dimension.")

    kdes = []
    for trial in range(len(con_sim_data)):
        s = con_sim_data[trial]

        # Compute a separate KDE for each combination of categorical variables.
        dens_u = {}
        for category in categories:
            # Get the subset of simulations that correspond to this category
            dsub = s[cat_sim_data[trial] == category]

            # If we didn't get enough simulation results for this category, don't do
            # a KDE
            if len(dsub) < 10:
                dens_u[category] = (None, None)
                continue

            # If any dimension of the data has a 0 range (all are same value) then
            # this will cause problems doing the KDE, skip.
            data_range = (
                np.max(dsub) - np.min(dsub)
                if dsub.ndim == 1
                else np.amax(dsub, 1) - np.amin(dsub, 1)
            )
            if np.any(data_range == 0):
                dens_u[category] = (None, None)
                warnings.warn(
                    BadLikelihoodWarning(
                        f"Could not perform kernel density estimate. Range of simulation data was 0 for at least "
                        f"one dimension. Range={data_range}"
                    )
                )
                continue

            # Do KDE
            fKDE = fastKDE.fastKDE(dsub, do_save_marginals=False)

            pdf = fKDE.pdf
            axes = fKDE.axes

            # Scale the pdf by the fraction of simulations that fall within this category
            pdf = pdf * (len(dsub) / len(s))

            # Save the KDE values and axes for this category
            dens_u[category] = (pdf, axes)

        kdes.append(dens_u)

    # If we are passed experimental data, evaluate the KDE at the experimental data points
    if exp_data is not None:
        # For numerical reasons, make zero probability a really small value. This is because we are taking logs
        # of the probabilities at the end.
        ZERO_PROB = 1e-10

        kdes_eval = np.zeros((len(exp_data),))
        for trial in range(len(exp_data)):
            # Extract the categorical values for this experimental trial
            exp_trial_cat = exp_data[trial, categorical_dims]

            if len(exp_trial_cat) == 1:
                exp_trial_cat = exp_trial_cat[0]

            # Get the right KDE for this trial, if all simulation trials have been combined
            # use that KDE for all trials of experimental data.
            if len(kdes) == 1:
                kde, axes = kdes[0].get(exp_trial_cat, (None, None))
            else:
                kde, axes = kdes[trial].get(exp_trial_cat, (None, None))

            # Linear interpolation using the grid we computed the KDE
            # on.
            if kde is not None:
                kdes_eval[trial] = interpn(
                    axes,
                    kde,
                    exp_data[trial, ~categorical_dims],
                    method="linear",
                    bounds_error=False,
                    fill_value=ZERO_PROB,
                )
            else:
                kdes_eval[trial] = ZERO_PROB

        # Check to see if any of the trials have non-zero likelihood, if not, something is probably wrong
        # and we should warn the user.
        if all(kdes_eval == ZERO_PROB):
            warnings.warn(
                BadLikelihoodWarning(
                    "Evaluating likelihood generated by simulation data resulted in zero values for all trials "
                    "of experimental data. This means the model is not generating data similar to your "
                    "experimental data. If you have categorical dimensions, make sure values match exactly to "
                    "output values of the composition. Also make sure parameter ranges you are searching over "
                    "are reasonable for your data."
                )
            )

        # Make 0 densities very small so log doesn't explode later
        kdes_eval[kdes_eval == 0.0] = ZERO_PROB

        return kdes_eval

    else:
        return kdes


class PECOptimizationFunction(OptimizationFunction):
    """
    A subclass of OptimizationFunction that is used to interface with the PEC. This class is used to specify the
    search method to utilize for optimization or data fitting. It is not to be confused with the `objective_function`
    that defines the optimization problem to be solved.

    Arguments
    ---------

    method :
        The search method to use for optimization. The following methods are currently supported:

            - 'differential_evolution' : Differential evolution as implemented by scipy.optimize.differential_evolution
            - optuna.samplers: Pass any instance of an optuna sampler to use optuna for optimization.
            - Type[optuna.samplers.BaseSampler]: Pass a class of type optuna.samplers.BaseSampler to use optuna
            for optimization. In this case, the random seed used for the sampler will be the same as the seed used
            as the intial_seed passed to PEC at contruction. Additonal desired keyword arguments can be passed to the
            sampler via the optuna_kwargs argument.

    optuna_kwargs :
        A dictionary of keyword arguments to pass to the optuna sampler. This is only used if method is an class of
        type optuna.samplers.BaseSampler. Note: this argument is ignored if method is an already instantiated instance
        of an optuna sampler.

    objective_function :
        The objective function to use for optimization. This is the function that defines the optimization problem the
        PEC is trying to solve. The function is used to evaluate the `values <Mechanism_Base.value>` of the
        `outcome_variables <ParameterEstimationComposition.outcome_variables>`, according to which combinations of
        `parameters <ParameterEstimationComposition.parameters>` are assessed; this must be an `Callable`
        that takes a 3d array as its only argument, the shape of which must be (**num_estimates**, **num_trials**,
        number of **outcome_variables**).  The function should specify how to aggregate the value of each
        **outcome_variable** over **num_estimates** and/or **num_trials** if either is greater than 1.

    max_iterations :
        The maximum number of iterations to run the optimization for. In differential evolution, this is the number of
        generations. In optuna, this is the number of trials.

    direction :
        Whether to maximize or minimize the objective function. If 'maximize', the objective function is maximized. If
        'minimize', the objective function is minimized.


    """
    class Parameters(OptimizationFunction.Parameters):
        initial_seed = SharedParameter(attribute_name='owner')

    @check_user_specified
    @beartype
    def __init__(
        self,
        method: Union[Literal["differential_evolution"], optuna.samplers.BaseSampler, Type[optuna.samplers.BaseSampler]],
        optuna_kwargs: Optional[Mapping] = None,
        objective_function: Optional[Callable] = None,
        search_space=None,
        save_samples: Optional[bool] = None,
        save_values: Optional[bool] = None,
        max_iterations: int = 500,
        direction: Literal["maximize", "minimize"] = "maximize",
        **kwargs,
    ):
        self.method = method
        self._optuna_kwargs = {} if optuna_kwargs is None else {**optuna_kwargs}

        self.direction = direction

        # The outcome variables to select from the composition's output need to be specified. These will be
        # set automatically by the PEC when PECOptimizationFunction is passed to it.
        self.outcome_variable_indices = None

        # The objective function to use for optimization. We can't set objective_function directly
        # because that will be set to agent_rep.evaluate when the PECOptimizationFunction is passed to
        # the OCM. Instead, we set self._pec_objective_function to the objective function, self.objective_function
        # will be used to compute just the simulation results, the these will then be passed to the
        # _pec_objective_function. Very confusing!
        self._pec_objective_function = objective_function

        # Are we in data fitting mode, or generic optimization. This is set automatically by the PEC when
        # PECOptimizationFunction is passed to it. It only really determines whether some cosmetic
        # things.
        self.data_fitting_mode = False

        # This is a bit confusing but PECOptimizationFunction utilizes the OCM search machinery only to run
        # simulations of the composition under different randomization. Thus, regardless of the method passed
        # to PECOptimize, we always set the search_function and search_termination_function for GridSearch.
        # The grid in our case is only over the randomization control signal.
        search_function = self._traverse_grid
        search_termination_function = self._grid_complete

        # When the OCM passes in the search space, we need to modify it so that the fitting parameters are
        # set to single values since we want to use SciPy optimize to drive the search for these parameters.
        # The randomization control signal is not set to a single value so that the composition still uses
        # the evaluate machinery to get the different simulations for a given setting of parameters chosen
        # by scipy during optimization. This variable keeps track of the original search space.
        self._full_search_space = None

        # Set num_iterations to a default value of 1, this will be reset in reset() based on the search space
        self.num_iterations = 1

        # Store max_iterations, this should be a common parameter for all optimization methods
        self.max_iterations = max_iterations

        # A cached copy of our log-likelihood function. This can only be created after the function has been assigned
        # to a OptimizationControlMechanism under and ParameterEstimationComposition.
        self._ll_func = None

        # This is the generation number we are on in the search, this corresponds to iterations in
        # differential_evolution
        self.gen_count = 1

        # Keeps track of the number of objective function evaluations during search
        self.num_evals = 0

        # Keep track of the best parameters
        self._best_params = {}

        self._method_kwargs = kwargs if kwargs else {}

        super().__init__(
            search_space=search_space,
            save_samples=save_samples,
            save_values=save_values,
            search_function=search_function,
            search_termination_function=search_termination_function,
            aggregation_function=None,
        )

    def set_pec_objective_function(self, objective_function: Callable):
        """
        Set the PEC objective function, this is the function that will be called by the OCM to evaluate
        the simulation results generated by the composition when it is simulated by the PEC.
        """
        self._pec_objective_function = objective_function

    @handle_external_context(fallback_most_recent=True)
    def reset(self, search_space, context=None, **kwargs):
        """Assign size of `search_space <MaxLikelihoodEstimator.search_space>"""

        # We need to modify the search space
        self._full_search_space = copy.deepcopy(search_space)

        # Modify all of the search space (except the randomization control signal) so that with each
        # call to evaluate we only evaluate a single parameter setting. Scipy optimize will direct
        # the search procedure so we will reset the actual value of these singleton iterators dynamically
        # on each search iteration executed during the call to _function.
        randomization_dimension = kwargs.get(
            "randomization_dimension", len(search_space) - 1
        )
        for i in range(len(search_space)):
            if i != randomization_dimension:
                search_space[i] = SampleIterator([next(search_space[i])])

        super().reset(search_space=search_space, context=context, **kwargs)
        owner_str = ""
        if self.owner:
            owner_str = f" of {self.owner.name}"
        for i in search_space:
            if i is None:
                raise OptimizationFunctionError(
                    f"Invalid {repr(SEARCH_SPACE)} arg for {self.name}{owner_str}; "
                    f"every dimension must be assigned a {SampleIterator.__name__}."
                )
            if i.num is None:
                raise OptimizationFunctionError(
                    f"Invalid {repr(SEARCH_SPACE)} arg for {self.name}{owner_str}; each "
                    f"{SampleIterator.__name__} must have a value for its 'num' attribute."
                )

        self.num_iterations = np.prod([i.num for i in search_space])

    def _run_simulations(self, *args, context=None):
        """
        Run the simulations we need for estimating the likelihood for given control allocation.
        This function has side effects as it sets the search_space parameter of the
        OptimizationFunction to the control allocation.
        """

        # Use the default variable for the function (control allocation), we don't need this for data fitting.
        variable = self.defaults.variable

        # Check that we have the proper number of arguments to map to the fitting parameters.
        if len(args) != len(self.fit_param_names):
            raise ValueError(
                f"Expected {len(self.fit_param_names)} arguments, got {len(args)}"
            )

        # If the model is in the inputs, then inputs are passed as list of lists and we need to add the fitting
        # parameters to each trial as a concatenated list
        inputs = self.owner.composition.controller._pec_input_values
        self.owner.composition.controller.set_parameters_in_inputs(parameters=args, inputs=inputs)

        # Reset the search grid
        self.reset_grid(context)

        # Evaluate objective_function for each sample
        last_sample, last_value, all_samples, all_values = self._evaluate(
            variable=variable,
            context=context,
            params=None,
            fit_evaluate=True,
        )

        # We need to swap the simulation (randomization dimension) with the output dimension so things
        # are in the right order passing to the objective_function call signature.
        all_values = np.transpose(all_values, (0, 2, 1))

        return all_values

    def _make_objective_func(self, context=None):
        """
        Make an objective function to pass to an optimization algorithm. Creates a function that runs simulations and
        then feeds the results self._pec_objective_func. This cannot be invoked until the PECOptimizationFunction
        (self) has been assigned to an OptimizationControlMechanism.
        """

        def objfunc(*args):
            sim_data = self._run_simulations(*args, context=context)

            # The composition might have more outputs than outcome variables, we need to subset the ones we need.
            sim_data = sim_data[:, :, self.outcome_variable_indices]

            return self._pec_objective_function(sim_data)

        return objfunc

    def _function(self, variable=None, context=None, params=None, **kwargs):
        """
        Run the optimization algorithm to find the optimal control allocation.
        """

        optimal_sample = self.variable
        optimal_value = np.array([1.0])
        saved_samples = []
        saved_values = []

        if not self.is_initializing:
            ocm = self.owner
            if ocm is None:
                raise ValueError(
                    "PECOptimizationFunction must be assigned to an OptimizationControlMechanism, "
                    "self.owner is None"
                )

            # Get the objective function that we are trying to minimize
            f = self._make_objective_func(context=context)

            # Run the MLE optimization
            results = self._fit(obj_func=f, context=context)

            # Get the optimal function value and sample
            optimal_value = results["optimal_value"]

            optimal_sample = list(results["fitted_params"].values())

            # Replace randomization dimension to match expected dimension of output_values of OCM. This is ugly but
            # necessary.
            if self.owner.num_estimates is not None:
                optimal_sample = optimal_sample + [0.0]

        return optimal_sample, optimal_value, saved_samples, saved_values

    @property
    def obj_func_desc_str(self):
        return "Log-Likelihood" if self.data_fitting_mode else "Obj-Func-Value"

    @property
    def opt_task_desc_str(self):
        direction_str = "Maximizing" if self.direction == "maximize" else "Minimizing"
        task_disp = (
            "Maximum Likelihood Estimation"
            if self.data_fitting_mode
            else f"{direction_str} Objective Function"
        )
        return f"{task_disp} (num_estimates={self.num_estimates}) ..."

    def _fit(
        self,
        obj_func: Callable,
        display_iter: bool = True,
        context: Context = None,
    ):
        if self.method == "differential_evolution":
            return self._fit_differential_evolution(obj_func, display_iter, context)
        elif isinstance(self.method, optuna.samplers.BaseSampler):

            if self.owner.initial_seed is not None:
                warnings.warn("initial_seed on PEC is not None, but instantiated optuna sampler is being used. If you "
                              "want deterministic behavior, make sure to specify seed on optuna sampler as well")

            return self._fit_optuna(
                obj_func=obj_func, opt_func=self.method, display_iter=display_iter
            )
        # If this is a class of type base sampler, instantiate it and pass it to _fit_optuna
        elif isinstance(self.method, type) and issubclass(self.method, optuna.samplers.BaseSampler):

            if self.owner.initial_seed is not None:
                if "seed" in self._optuna_kwargs:
                    warnings.warn(
                        f"Overriding seed passed to optuna sampler with seed passed to PEC. "
                        f"Optuna sampler seed: {self._optuna_kwargs['seed']}, PEC.initial_seed: {self.owner.initial_seed}"
                    )

                self._optuna_kwargs["seed"] = self.owner.initial_seed

            return self._fit_optuna(
                obj_func=obj_func, opt_func=self.method(**self._optuna_kwargs), display_iter=display_iter
            )
        else:
            raise ValueError(f"Invalid optimization_function method: {self.method}")

    def _make_obj_func_wrapper(
        self,
        progress,
        display_iter,
        warns,
        warns_with_params,
        obj_func,
        like_eval_task=None,
        evals_per_iteration=None,
        ignore_direction=False,
    ):
        """
        Create a wrapper function for the objective function that keeps track of progress and warnings.
        """
        direction = 1 if self.direction == "minimize" or ignore_direction else -1

        # This is the number of evaluations we need per search iteration.
        self.num_evals = 0

        # Create a wrapper function for the objective. This lets us keep track of progress and such
        def objfunc_wrapper(x):
            params = dict(zip(self.fit_param_names, x))
            t0 = time.time()
            obj_val = obj_func(*x)
            p = direction * obj_val
            elapsed = time.time() - t0
            self.num_evals = self.num_evals + 1

            # Keep a log of warnings and the parameters that caused them
            if len(warns) > 0 and warns[-1].category == PECObjectiveFuncWarning:
                warns_with_params.append((warns[-1], params))

            # Are we displaying each iteration
            if display_iter:
                # If we got a warning generating the objective function value, report it
                if len(warns) > 0 and warns[-1].category == PECObjectiveFuncWarning:
                    progress.console.print(f"Warning: ", style="bold red")
                    progress.console.print(f"{warns[-1].message}", style="bold red")
                    progress.console.print(
                        f"{get_param_str(params)}, {self.obj_func_desc_str}: {obj_val}, "
                        f"Eval-Time: {elapsed} (seconds)",
                        style="bold red",
                        highlight=False,
                    )
                    # Clear the warnings
                    warns.clear()
                else:
                    progress.console.print(
                        f"{get_param_str(params)}, {self.obj_func_desc_str}: {obj_val}, "
                        f"Eval-Time: {elapsed} (seconds)",
                        highlight=False,
                    )

                # Certain algorithms like differential evolution evaluate the objective function multiple times per
                # iteration. We need to update the progress bar accordingly.
                if evals_per_iteration is not None:
                    # We need to update the progress bar differently depending on whether we are doing
                    # the first generation (which is twice as long) or not.
                    if self.num_evals < 2 * evals_per_iteration:
                        max_evals = 2 * evals_per_iteration
                        progress.tasks[like_eval_task].total = max_evals
                        eval_task_str = f"|-- Iteration {self.gen_count} ..."
                        progress.tasks[like_eval_task].description = eval_task_str
                        progress.update(
                            like_eval_task, completed=self.num_evals % max_evals
                        )
                    else:
                        max_evals = evals_per_iteration
                        progress.tasks[like_eval_task].total = max_evals
                        eval_task_str = f"|-- Iteration {self.gen_count} ..."
                        progress.tasks[like_eval_task].description = eval_task_str
                        progress.update(
                            like_eval_task,
                            completed=(self.num_evals - (2 * evals_per_iteration))
                            % max_evals,
                        )

            return p

        return objfunc_wrapper

    def _fit_differential_evolution(
        self,
        obj_func: Callable,
        display_iter: bool = True,
        context: Context = None,
    ):
        """
        Implementation of search using scipy's differential_evolution algorithm.
        """

        bounds = self.fit_param_bounds

        # We just need the upper and lower bounds for the differential evolution algorithm. The step size is not used.
        bounds = list([(lb, ub) for name, (lb, ub, step) in bounds.items()])

        # Get a seed to pass to scipy for its search. Make this dependent on the seed of the
        # OCM
        seed_for_scipy = self._get_current_parameter_value('initial_seed', context)
        seed_for_scipy = try_extract_0d_array_item(seed_for_scipy)

        direction = 1 if self.direction == "minimize" else -1

        with Progress(
            "[progress.description]{task.description}",
            BarColumn(),
            "Completed: [progress.percentage]{task.percentage:>3.0f}%",
            TimeRemainingColumn(),
        ) as progress:
            opt_task = progress.add_task(self.opt_task_desc_str, total=100)

            # This is the number of evaluations we need per search iteration.
            evals_per_iteration = 15 * len(self.fit_param_names)
            self.gen_count = 1

            if display_iter:
                eval_task_str = f"|-- Iteration 1 ..."
                like_eval_task = progress.add_task(
                    eval_task_str, total=evals_per_iteration
                )

            progress.update(opt_task, completed=0)

            warns_with_params = []
            with warnings.catch_warnings(record=True) as warns:
                objfunc_wrapper = self._make_obj_func_wrapper(
                    progress,
                    display_iter,
                    warns,
                    warns_with_params,
                    obj_func,
                    like_eval_task,
                    evals_per_iteration,
                )

                def progress_callback(x, convergence):
                    params = dict(zip(self.fit_param_names, x))
                    convergence_pct = 100.0 * convergence
                    progress.console.print(
                        f"[green]Current Best Parameters: {get_param_str(params)}, "
                        f"{self.obj_func_desc_str}: {obj_func(*x)}, "
                        f"Convergence: {convergence_pct}"
                    )

                    # If we encounter any PECObjectiveFuncWarnings. Summarize them for the user
                    if len(warns_with_params) > 0:
                        progress.console.print(
                            f"Warning: degenerate {self.obj_func_desc_str} values for the following parameter values ",
                            style="bold red",
                        )
                        for w in warns_with_params:
                            progress.console.print(
                                f"\t{get_param_str(w[1])}", style="bold red"
                            )
                        progress.console.print(
                            "If these warnings are intermittent, check to see if search "
                            "space is appropriately bounded. If they are constant, and you are fitting to"
                            "data, make sure experimental data and output of your composition are similar.",
                            style="bold red",
                        )

                    progress.update(opt_task, completed=convergence_pct)
                    self.gen_count = self.gen_count + 1

                r = differential_evolution(
                    objfunc_wrapper,
                    bounds,
                    callback=progress_callback,
                    maxiter=self.parameters.max_iterations.get() - 1,
                    seed=seed_for_scipy,
                    popsize=15,
                    polish=False,
                    **self._method_kwargs
                )

            # Bind the fitted parameters to their names
            fitted_params = dict(zip(list(self.fit_param_names), r.x))

        # Save all the results
        output_dict = {
            "fitted_params": fitted_params,
            "optimal_value": direction * r.fun,
        }

        return output_dict

    def _fit_optuna(
        self,
        obj_func: Callable,
        opt_func: Union[optuna.samplers.BaseSampler, Type[optuna.samplers.BaseSampler]],
        display_iter: bool = True,
    ):
        with Progress(
            "[progress.description]{task.description}",
            BarColumn(),
            "Completed: [progress.percentage]{task.percentage:>3.0f}%",
            TimeRemainingColumn(),
        ) as progress:
            max_iterations = self.parameters.max_iterations.get()

            opt_task = progress.add_task(self.opt_task_desc_str, total=max_iterations)
            progress.update(opt_task, completed=0)

            warns_with_params = []
            with warnings.catch_warnings(record=True) as warns:
                # Create a wrapper for the objective function that will let us catch warnings and record progress.
                # For optuna, we can ignore the direction of search because it is handled by the Optuna API when
                # setting up the optimization.
                objfunc_wrapper = self._make_obj_func_wrapper(
                    progress=progress,
                    display_iter=display_iter,
                    warns=warns,
                    warns_with_params=warns_with_params,
                    obj_func=obj_func,
                    ignore_direction=True,
                )

                # Optuna has an interface where the objective function calls the API to get
                # the current values for the parameter rather than them being passed
                # directly. So we need to wrap the wrapper
                def objfunc_wrapper_wrapper(trial):
                    for name, (lower, upper, step) in self.fit_param_bounds.items():
                        trial.suggest_float(name, lower, upper, step=step)

                    return objfunc_wrapper(list(trial.params.values()))

                self._best_params = {}

                def progress_callback(study, trial):
                    if self._best_params != study.best_params:
                        self._best_params = study.best_params
                        progress.console.print(
                            f"[green]Current Best Parameters: {get_param_str(self._best_params)}, "
                            f"{self.obj_func_desc_str}: {study.best_value}, "
                        )

                    # If we encounter any PECObjectiveFuncWarnings. Summarize them for the user
                    if len(warns_with_params) > 0:
                        progress.console.print(
                            f"Warning: degenerate {self.obj_func_desc_str} values for the following parameter values ",
                            style="bold red",
                        )
                        for w in warns_with_params:
                            progress.console.print(
                                f"\t{get_param_str(w[1])}", style="bold red"
                            )
                        progress.console.print(
                            "If these warnings are intermittent, check to see if search "
                            "space is appropriately bounded. If they are constant, and you are fitting to"
                            "data, make sure experimental data and output of your composition are similar.",
                            style="bold red",
                        )

                    progress.update(opt_task, advance=1)

                # Turn off optuna logging except for errors or warnings, it doesn't work well with our PNL progress bar
                optuna.logging.set_verbosity(optuna.logging.WARNING)

                study = optuna.create_study(
                    sampler=opt_func, direction=self.direction
                )
                study.optimize(
                    objfunc_wrapper_wrapper,
                    n_trials=max_iterations,
                    callbacks=[progress_callback],
                )

            # Bind the fitted parameters to their names
            fitted_params = dict(
                zip(list(self.fit_param_names), study.best_params.values())
            )

        # Save all the results
        output_dict = {
            "fitted_params": fitted_params,
            "optimal_value": study.best_value,
        }

        return output_dict

    @property
    def fit_param_names(self) -> List[str]:
        """Get a unique name for each parameter in the fit."""
        if self.owner is not None:
            # Go through each parameter and create a unique name for it
            if not self.owner.depends_on:
                return [f"{mech.name}.{param_name}"
                        for param_name, mech in self.owner.fit_parameters.keys()]
            else:
                names = []
                for param_name, mech in self.owner.fit_parameters.keys():
                    if (param_name, mech) in self.owner.cond_levels:
                        for level in self.owner.cond_levels[(param_name, mech)]:
                            names.append(f"{mech.name}.{param_name}[{level}]")
                    else:
                        names.append(f"{mech.name}.{param_name}")

                return names

        else:
            return None

    @property
    def fit_param_bounds(self) -> Dict[str, Tuple[float, float, float]]:
        """
        Get the allocation samples for just the fitting parameters. Whatever they are, turn them into upper and lower
        bounds, with a step size as well.

        Returns:
            A dict mapping parameter names to (lower, upper) bounds.
            A dict mapping parameter names to step sizes.
        """

        if self.owner is not None:

            if not self.owner.depends_on:
                bounds = [(float(min(s)), float(max(s))) for s in self.owner.fit_parameters.values()]
                steps = [np.unique(np.diff(s).round(decimals=5)) for s in self.owner.fit_parameters.values()]
            else:
                bounds = []
                steps = []
                for param_name, mech in self.owner.fit_parameters.keys():
                    s = self.owner.fit_parameters[(param_name, mech)]
                    if (param_name, mech) in self.owner.cond_levels:
                        for _ in self.owner.cond_levels[(param_name, mech)]:
                            bounds.append((float(min(s)), float(max(s))))
                            steps.append(np.unique(np.diff(s).round(decimals=5)))
                    else:
                        bounds.append((float(min(s)), float(max(s))))
                        steps.append(np.unique(np.diff(s).round(decimals=5)))

            # We also check if step size is constant, if not we raise an error
            for s in steps:
                if len(s) > 1:
                    raise ValueError("Step size for each parameter must be constant")

            steps = [float(s[0]) for s in steps]

            return dict(
                zip(
                    self.fit_param_names,
                    ((l, u, s) for (l, u), s in zip(bounds, steps)),
                )
            )
        else:
            return None

    @handle_external_context(fallback_most_recent=True)
    def log_likelihood(self, *args, context=None):
        """
        Compute the log-likelihood of the data given the specified parameters of the model. This function will raise
        aa exception if the function has not been assigned as the function of and OptimizationControlMechanism. An
        OCM is required in order to simulate results of the model for computing the likelihood.

        Arguments
        ---------
        *args :
            Positional args, one for each paramter of the model. These must correspond directly to the parameters that
            have been specified in the `parameters` argument of the constructor.

        context: Context
            The context in which the log-likelihood is to be evaluated.

        Returns
        -------
        The sum of the log-likelihoods of the data given the specified parameters of the model.
        """

        if self.owner is None:
            raise ValueError(
                "Cannot compute a log-likelihood without being assigned as the function of an "
                "OptimizationControlMechanism. See the documentation for the "
                "ParameterEstimationControlMechanism for more information."
            )

        # Make sure we have instantiated the log-likelihood function.
        if self._ll_func is None:
            self._ll_func = self._make_objective_func(context=context)

        context.execution_phase = ContextFlags.PROCESSING
        ll, sim_data = self._ll_func(*args)
        context.remove_flag(ContextFlags.PROCESSING)

        return ll, sim_data
